#Set-StrictMode -Version 2.0

<#
.SYNOPSIS
    Deploys a new version of a JCMS war into an existing exploded webapp.
	
.DESCRIPTION
    Extract the content of the specified WAR and deploys all the applicative
	parts while keeping the existing data. Optionnaly merging store of
	new and existing webapp.
	The existing webapp directory is always backuped.
	
	Requirements : 
	
		* PowerShell V2
		
		* Administrator privileges : This script recreate symbolink lync
		in the new webapp directory, administrator privileges are required
		to perform this operation.		
		
		* Symbolink links : The following directories of the existing 
		exploded webapp MUST be symbolink links referencing directories
		outside the webapp directory : 
			archives\
			upload\
			WEB-INF\data\
		
		* Java : a Java Virtual Machine must be installed and available
		in the current execution PATH if you want to merge store.xml files.
		
.PARAMETER WarPath
    The new JCMS war to be deployed. Must be a valid Web Archive file.
.PARAMETER TargetDir
    The existing exploded webapp in which to deploy the new war.
.PARAMETER StoreAction
	The action to perform on store.xml file, from the following :
		"keep" (default) : keep store.xml of existing webapp 
		"overwrite"      : overwrite existing store with store from war
		"merge"          : merge existing and new store 
	existing store is backuped when using 'overwrite' or 'merge' 
.PARAMETER BackupDirectory
	A backup directory : old webapp will be moved inside this directory
	once deployement has been performed. Default behavior is to leave
	webapp in same directory and simply rename the webapp directory.
.PARAMETER DryRun
	Do not perform any modification on the file system if this switch is enabled.


.EXAMPLE
    C:\PS> Deploy-JCMS c:\deploy\newwebapp.war c:\tomcat\webapps\ROOT
    Backup the existing webapp in ROOT directory and deploys the newwebapp.war.
	The existing store.xml file kept un-modified

.EXAMPLE
    C:\PS> Deploy-JCMS -StoreAction "merge" newjcms.war c:\mywebapp
    Deploys the newjcms.war in the mywebapp directory.
	The existing store.xml is merged with the store.xml present in the nexjcms.war

.NOTES
    Author: Jalios
#>
param(
	$WarPath = $null,
	$TargetDir = $null,
	$StoreAction = "keep",
	$BackupDirectory = $null,
	$LogMessage = $null,
	
	[switch] $Help = $false,
	[switch] $DryRun = $false, # == should be implemented using "WhatIf" ?
	[switch] $Verbose = $false,
	[switch] $Debug = $false
)

$script:deployps1_revision = [regex]::Replace('$Revision: 61537 $', '^\$\w+:\s*([0-9]*)\s*\$$', '$1') ;
$script:deployps1_date = [regex]::Replace('$Date: 2013-08-30 13:58:42 +0200 (ven., 30 août 2013) $', '^\$\w+:\s*(.*)\s*\$$', '$1') ;

function Write-Usage([string]$msg, [System.Management.Automation.ErrorCategory]$category)
{
	if ($msg) { 
		if ($category) {
			Write-Error -Category $category $msg
		} else {
			Write-Error $msg
		}
	}

	Write-Host @"
PS > &"deploy.ps1" [-BackupDirectory dir] [-WarPath] newwebapp.war [-TargetDir] targetdir 
-help                  display version and this help message and exit

-StoreAction action    action to perform on store.xml file, from the following :
				"keep" (default) : keep store.xml of existing webapp 
				"overwrite"      : overwrite existing store with store from war
				"merge"          : merge existing and new store 
				existing store is backuped when using "overwrite" or "merge"
				
-BackupDirectory dir   backup directory : old webapp will be moved inside this directory
				once deployement has been performed. Default behavior is to leave
				webapp in same directory and simply rename the webapp directory.
-LogMessage message    An optionnal message logged in WEB-INF\data\logs\deploy.log
				upon deployement success
## TODO nodiff option
-NoDiff       Do not log webapp differences after deployement (faster)
-Verbose      verbose mode, display very detailed operation 
-DryRun       dryrun, do not perform any filesystem modification but display what 
              would have been done"

"@
exit 1;
}

function CheckLastExitStatus() {
	$lastCommandExitStatus = $LASTEXITCODE ;
	$lastError = $error[0] ;
	if ( $lastError -or ($lastCommandExitStatus -and $lastCommandExitStatus -ne 0))
	{
		Write-Verbose "lastCommandExitStatus : $lastCommandExitStatus " ;
		Write-Verbose "lastError : $lastError " ;
		Write-Error "Last command could not be executed correctly. Aborting" ;
		Rollback ;
		exit $lastCommandExitStatus ;
	}
	
}

#######################################
# Rollback methods

function Setup-RollbackScript() {  
 	$script:RollbackScript = "$TmpDir/rollback.ps1" ;
 	$script:RollbackScriptTmp = "$TmpDir/rollback.ps1.tmp" ;
	if (!$DryRun) {
		$script:RollbackScript = New-Item -Type file -Force $script:RollbackScript ; 
		$script:RollbackScriptTmp = New-Item -Type file -Force $script:RollbackScriptTmp ; 
  		Write-Output "" | Out-File $RollbackScript;
  		Write-Output "" | Out-File $RollbackScriptTmp;
	}
}

# Add the specified command to the list of commands to be executed if 
# any failure occurs during processing.
# IMPORTANT: commmand MUST be enclosed in quotes and SHOULD NOT contain single quote !
# eg : Push-RollbackCommand "Get-ChildItem -recurse `"$foobar`""
function Push-RollbackCommand($command) {
	Write-Verbose "Adding rollback command '$command'" ;

	if (!$DryRun) {
		# Command must be executed in reverse order (last in, first out)
		# therefore we prepend the command using an intermediate file.
		# Also invoke the 'checkLastRollbackExitStatus' method to check it
		# the command was successful
		Write-Output "Write-Output '# $command'" | Out-File $RollbackScriptTmp ; CheckLastExitStatus ;
		Write-Output "$command ; " | Out-File -Append $RollbackScriptTmp ; CheckLastExitStatus ;
		Write-Output "Check-LastRollbackExitStatus ; `n`n "  | Out-File -Append $RollbackScriptTmp ; CheckLastExitStatus ;

		# Add previous commands (existing content of rollback script)
		Get-Content $RollbackScript | Out-File -Append $RollbackScriptTmp ; CheckLastExitStatus ;
		Move-Item -Force $RollbackScriptTmp $RollbackScript ; CheckLastExitStatus ;
	}
}
  
function Rollback() {
	if ($rollbackAsked) {
		return ;
	}
	$script:rollbackAsked = $true ;

	Write-Output "Deployement could not be completed : rollback all modifications ..."
	if ($RollbackScript -and (Test-Path -PathType Leaf $RollbackScript)) {
  
		# Prepend required methods
		Write-Output '
    
function Check-LastRollbackExitStatus() {
	$lastRollbackCommandExitStatus = $LASTEXITCODE ;
	$lastError = $error[0] ;
	if ($lastError -or ($lastRollbackCommandExitStatus -ne 0) ) {
		exit $lastRollbackCommandExitStatus ;
	}
}
' | Out-File $RollbackScriptTmp ;
		Get-Content $RollbackScript | Out-File -Append $RollbackScriptTmp ;
		Move-Item -Force $RollbackScriptTmp $RollbackScript ;

		# Log script content
		if ( $Verbose ) {
			Write-Verbose "Executing rollback script '$RollbackScript' :" ;
			Write-Verbose "# ----------------------------------------------------------" ;
			Write-Verbose "$( Get-Content $RollbackScript )" ;
			Write-Verbose "# ----------------------------------------------------------`n" ;
		}

		# Execute
		$error.Clear();
		& "$RollbackScript" ;
		$rollbackExitStatus = $LASTEXITCODE ;
		$roolbackError = $error[0]
		if ($roolbackError -or ($rollbackExitStatus -ne 0)) {
		  Write-Error 'ERROR !! Rollback could not be completed correctly. CHECK WEBAPP INTEGRITY !!' ;
		  exit $rollbackExitStatus ;
		}
	};   
	Write-Output "All modifications were rollbacked successfully."
}

#######################################
# Methods used to check requirements

function Check-Requirements() 
{
	if ($PSVersionTable["PSVersion"].Major -lt 2) {
		Write-Error "PowerShell v2 is required to run this script" ;
		exit 1;
	}

	Check-IsAdministrator ;

	# Execution environement
	$ScriptDirectoryPath = Split-Path $MyInvocation.ScriptName
	# unzip 
	$IonicAssemblyPath =  "$ScriptDirectoryPath\Ionic.Zip.dll";
	$IonicAssembly = [System.Reflection.Assembly]::LoadFrom($IonicAssemblyPath) ;
	if (!$IonicAssembly) {
		exit 1;
	}
	Write-Debug $IonicAssembly ;
	# ReparsePoint
	$ReparsePointsPath = "$ScriptDirectoryPath\ReparsePoints.dll";
	$ReparsePointsAssembly = [System.Reflection.Assembly]::LoadFrom($ReparsePointsPath) ;
	if (!$ReparsePointsAssembly) {
		exit 1;
	}
	Write-Debug $ReparsePointsAssembly ;
	# java
	java -version 
	if (!$?) {
		Write-Error "Java is required but could not be found" ;
		exit 1;
	}

	# Arguments 
  	if (($StoreAction -ne "keep") -and ($StoreAction -ne "overwrite") -and ($StoreAction -ne "merge"))
	{
		Write-Usage "Invalid -s store action '$StoreAction'. Valid actions : 'keep', 'overwrite', 'merge'" -Category InvalidArgument;
	}
	# Specified war exist and is a valid web archive
	$script:WarPath = Get-Item $WarPath ;
	Check-War -WarPath $WarPath
	
	# Target directory exists
	$script:TargetDir = Get-Item $TargetDir ;
	if (!($TargetDir)) {
		Write-Usage "TargetDir argument is required." -Category InvalidArgument;
	}
	
	if (!(Test-Path -PathType Container $TargetDir)) {
		Write-Usage "Target directory does not exist or is not a valid directory : '$TargetDir'" -Category InvalidData;
	}
	
	# Backup directory exists (if specified)
	if ($BackupDirectory) { 
		if (!(Test-Path -PathType Container "$BackupDirectory")) {
			Write-Usage "Invalid backup directory : '$BackupDirectory'" -Category InvalidData;
		}
		$script:BackupDirectory = Get-Item $BackupDirectory ; 
	}
	
	# Webapp directories uses symlink
	Check-Symlink "$TargetDir\archives" ;
	Check-Symlink "$TargetDir\upload" ;
	Check-Symlink "$TargetDir\WEB-INF\data" ;
	
	
	## TODO : Number of plugin in zip vs number of plugin in webapp
}

function Check-IsAdministrator
{
	$currentPrincipal = New-Object Security.Principal.WindowsPrincipal( [Security.Principal.WindowsIdentity]::GetCurrent() ) 
	if (!$currentPrincipal -or !$currentPrincipal.IsInRole( [Security.Principal.WindowsBuiltInRole]::Administrator )) {
		Write-Usage "Script must be run with administrator privileges" ; 
	}
}

function Check-War([string] $WarPath) 
{
	if (!($WarPath)) {
		Write-Usage "WarPath argument is required." -Category InvalidArgument ;
	}
	if (!(Test-Path -PathType Leaf $WarPath)) {
		Write-Usage "War file does not exist or is not a valid file: '$WarPath'" -Category InvalidData;
	}
	try {
		Write-Verbose "Reading war file '$WarPath'" ;
		$zipfile = [Ionic.Zip.ZipFile]::Read($WarPath);
		$webXml = $zipfile["WEB-INF\web.xml"];
		Write-Debug "web.xml zip entry : '$webXml'" ;
		$zipfile.Dispose();
		$zipfile = $null;
	} catch {
		Write-Usage "Invalid web archive (zip is invalid) : '$WarPath' $_" -Category InvalidData;
	}
	if ($null -eq $webXml)
	{
		Write-Usage "Invalid web archive (web.xml could not be found) : '$WarPath'" -Category InvalidData ;
	}
}

# Test if the specified path is a symlink
# http://stackoverflow.com/questions/817794/find-out-whether-a-file-is-a-symlink-in-powershell
function Test-Symlink([string] $path) {
	$file = Get-Item $path -Force -ea 0
	if ($file) {
		return [bool]($file.Attributes -band [IO.FileAttributes]::ReparsePoint)
	}
	return $false;
}

# mklink /D "C:\deploy\webapp\archives" "c:\deploy\webapp-data\archives"
# mklink /D "C:\deploy\webapp\upload" "c:\deploy\webapp-data\upload"
# mklink /D "C:\deploy\webapp\WEB-INF\data" "c:\deploy\webapp-data\WEB-INF\data"
function Check-Symlink([string] $DirectoryPath) 
{
	if (!(Test-Path -PathType Container -Path $directoryPath))
	{
		Write-Usage "'$DirectoryPath' MUST be a valid directory." ; 
	}
	if (!(Test-Symlink $directoryPath))
	{
		Write-Usage "'$DirectoryPath' MUST be a symbolic link." ; 
	}
}

#######################################
# Environement

# Define the file suffix used every time a copy is made of old data
function Setup-BackupSuffix() {
	$script:BackupSuffix = "$(Get-Date -format 'yyyyMMdd-hhmmss')" ;
}

# Create temporary work directory
function Setup-TmpDir() {
	$script:TmpDir = "$env:temp\JaliosJCMS-Deploy-$(Get-Date -format 'yyyyMMdd-HHmmss')" ;
	if (!$DryRun) {
		$script:TmpDir = New-Item -Type Directory $TmpDir ;
	}
}

# Build Java classpath required to execute java program
function Setup-JavaClassPath() {
	$script:JAVA_CLASSPATH = "$TargetDir\WEB-INF\classes" ;
	Get-ChildItem $TargetDir\WEB-INF\lib\*.jar |%{ 
		$script:JAVA_CLASSPATH += ";" + $_.Fullname ;
	}
}

function Setup-Environment() {
  Setup-BackupSuffix ;
  Setup-TmpDir ;
  Setup-JavaClassPath ;
  Setup-RollbackScript ;
}



#######################################
# JAVA methods

function Store-Merge($store1, $store2) {
  Write-Verbose "Executing java -classpath `"$JAVA_CLASSPATH`" com.jalios.jcms.tools.StoreMerge -output `"$store1`" `"$store1`" `"$store2`"" ;
  if ($DryRun) {
  	return;
  }
  
  
  java -classpath "$JAVA_CLASSPATH" com.jalios.jcms.tools.StoreMerge -output "$store1" "$store1" "$store2"
  switch ($LASTEXITCODE)
  {
		0 { # OK
			Write-Verbose "StoreMerge successful" ;
			return ;
		}
		1 { # Bad Parameters
			Write-Error "Invalid parameters specified to StoreMerge" ;
			Rollback ;
			exit 1 ; 
		}
		2 { # Error
			Write-Error "An error occured whild processing StoreMerge, check log" ;
			Rollback ;
			exit 2 ; 
		}
		3 { # Conflicts
			Write-Error "Conflicts were detected while processing StoreMerge" ;
			Rollback ;
			exit 3 ; 
		} 
		default { # unexpected error
			Write-Error "An unexpected error ($?) occured while invoking StoreMerge" ;
			exit $? ; 
		}
	}
}

#######################################
# main processing methods
# Unzip war in temp dir

function Unzip-War() {  
	$script:NewWebappDir = "$TmpDir\webapp" ;
	Write-Verbose "Unzipping war '$WarPath' to tmpdir '$NewWebappDir'" ;
	
	## Create NewWebappDir directory
	if (!$DryRun) {
		$script:NewWebappDir = New-Item $NewWebappDir -Type Directory ;
	}
	Write-Debug(" NewWebappDir    : '$script:NewWebappDir' - [$(Get-Type-FullName($script:NewWebappDir))]") ;
	
	## Unzip and Check process success
	Write-Verbose "Extract war file '$WarPath' to '$NewWebappDir'" ;
	try {
		$zipfile = [Ionic.Zip.ZipFile]::Read($WarPath);
		foreach ($zipentry in $zipfile)
		{
			# Write-Debug ("Extract " + $zipentry.FileName) ;
			if (!$DryRun) {
				$zipentry.Extract($NewWebappDir, [Ionic.Zip.ExtractExistingFileAction]::OverwriteSilently) ;
			}
		}
		$zipfile.Dispose();
	} catch [Exception] {
		Write-Error ("Could not extract Web Archive : " + $_) -Category WriteError ;
		if ($zipfile) { $zipfile.Dispose(); }
		$zipfile = $null;
		exit 1;
	}	
}
function Copy-SymLink([string]$RelativePath, [string]$BackupDirectoryName) {
	$LinkTarget = ([ReparsePoints.ReparsePoint]("$TargetDir\$RelativePath")).Target;
	if (!$LinkTarget) {
		Write-Error "Could not read symlink target for $TargetDir\$RelativePath" ;
		exit 1; ## TODO : cleanup war
	}
	
	Write-Verbose " - move $RelativePath directory (if any) \t: '$NewWebappDir\$RelativePath' to '$TmpDir\$BackupDirectoryName'" ;
	if (!$DryRun -and (Test-Path -PathType Container "$NewWebappDir\$RelativePath")) {
		Move-Item "$NewWebappDir\$RelativePath" "$TmpDir\$BackupDirectoryName" ; CheckLastExitStatus ;
	}
	
	Write-Verbose " - link $RelativePath directory \t: '$NewWebappDir\$RelativePath' --> '$LinkTarget'" ;
	if (!$DryRun) {
		cmd /c "mklink /d `"$NewWebappDir\$RelativePath`" `"$LinkTarget`"" ; CheckLastExitStatus ;
	}
}

function Copy-SymLinks() {
	Write-Verbose "Recreate symlinks in new webapp '$NewWebappDir'" ;
  
	Copy-SymLink "archives" "archives.original";
	Copy-SymLink "upload"  "upload.original";
	Copy-SymLink "WEB-INF\data" "data.original";
}

# Backup store.xml before overwrite or merge
function BackupStoreBeforeModification() {
	$script:backup_store_path = "$TargetDir\WEB-INF\data\store.xml.$BackupSuffix" ;
	Write-Output "[store.xml] Old 'store.xml' backuped to '$backup_store_path'" ;
	if (!$DryRun) {
		Copy-Item "$TargetDir\WEB-INF\data\store.xml" "$backup_store_path" ; CheckLastExitStatus ;
		Push-RollbackCommand "Move-Item -Force `"$backup_store_path`" `"$TargetDir\WEB-INF\data\store.xml`" ; " ; 
	}
}

# Process Store
function Process-Store() {
	Write-Verbose "Process store.xml" ;
  
	if ( "$StoreAction" -eq "keep" )
	{
		Write-Verbose "[store.xml] Keep existing store" ;
		# nothing to do
		Write-Output "[store.xml] store unmodified ('$TargetDir\WEB-INF\data\store.xml') " ;
	}
	
	elseif ( "$StoreAction" -eq "overwrite" )
	{
		Write-Verbose "[store.xml] Overwrite with new store : replace existing store.xml with new store.xml from WAR" ;
		BackupStoreBeforeModification ;
    
		# Overwrite webapp store with store.xml from WAR
		if (!$DryRun) { 
			Move-Item -Force "$TmpDir\data.original\store.xml" "$NewWebappDir\WEB-INF\data\store.xml"
		}
		Write-Output "[store.xml] store.xml retrieved from new war" ;
	}
    
	elseif ( "$StoreAction" -eq "merge" )
	{
		Write-Verbose "[store.xml] Merge stores" ;
		BackupStoreBeforeModification ;
    
		# Invoke Store Merge
		if (!$DryRun) { 
			Store-Merge  "$NewWebappDir\WEB-INF\data\store.xml" "$TmpDir\data.original\store.xml" ;
		}
		Write-Output "[store.xml] store.xml from existing webapp and from new war merged successfully " ;
	} 
}

# Process Types
function Process-Types() {
	Write-Verbose "Process types (replace existing types with new one)" ;

	# Backup existing types before using new one
	$script:backup_types_path="$TargetDir\WEB-INF\data\types.$BackupSuffix" ;
	Write-Output "[types] Old 'types' directory backuped to '$backup_types_path'" ;
	if (!$DryRun) {
		Move-Item "$NewWebappDir\WEB-INF\data\types" "$backup_types_path" ; CheckLastExitStatus ;
		Push-RollbackCommand "Move-Item `"$backup_types_path`" `"$NewWebappDir\WEB-INF\data\types`" ; " ; 
	}

	# Import Types from new webapp data
	if (!$DryRun) {
		Move-Item "$TmpDir\data.original\types" "$NewWebappDir\WEB-INF\data\types" ; CheckLastExitStatus ;
		Push-RollbackCommand "Move-Item `"$NewWebappDir\WEB-INF\data\types`" `"$TmpDir\data.original\types`" ; " ;
	}
}

# Process Workflow
function Process-Workflows() {
	Write-Verbose "Process workflows (replace existing worfklows with new one)" ;

	# Backup existing workflows before using new one
	$script:backup_workflows_path="$TargetDir\WEB-INF\data\workflows.$BackupSuffix" ;
	Write-Output "[workflows] Old 'workflows' directory backuped to '$backup_workflows_path'" ;
	if (!$DryRun) {
		Move-Item "$NewWebappDir\WEB-INF\data\workflows" "$backup_workflows_path" ; CheckLastExitStatus ;
		Push-RollbackCommand "Move-Item `"$backup_workflows_path`" `"$newWebappDir\WEB-INF\data\workflows`" ; " ; 
	}

	# Import Workflow from new webapp data
	if (!$DryRun) {
		Move-Item "$TmpDir\data.original\workflows" "$NewWebappDir\WEB-INF\data\workflows" ; CheckLastExitStatus ;
		Push-RollbackCommand "Move-Item `"$NewWebappDir\WEB-INF\data\workflows`" `"$TmpDir\data.original\workflows`"; " ;
	}
}


# Replace the target webapp with the newly modified webapp
function Process-Deploy() {
	Write-Verbose "Perform deployement (backup existing webapp and replace with new one)" ;

	# Backup existing webapp
	if ($BackupDirectory) {
		$script:backup_webapp_path="$BackupDirectory\$($TargetDir.name).$BackupSuffix" ;
	} else {
		$script:backup_webapp_path="$TargetDir.$BackupSuffix" ;
	}
	Write-Output "[webapp] Backup old webapp to '$backup_webapp_path'..." ;
	Write-Verbose "[webapp] move '$TargetDir' to '$backup_webapp_path'" ;
	if (!$DryRun) {
		Move-Item "$TargetDir" "$backup_webapp_path" ; 
		Write-Verbose "`$LASTEXITCODE : $LASTEXITCODE" ;
		Write-Verbose "`$_ : $_" ;
		
		CheckLastExitStatus ;
		Push-RollbackCommand "Move-Item `"$backup_webapp_path`" `"$TargetDir`" ; " ; 
	}
	Write-Output "[webapp] Old webapp backuped to '$backup_webapp_path'" ;

	# Move new webapp to target folder
	Write-Output "[webapp] Installing new webapp" ;
	Write-Verbose "[webapp] move '$NewWebappDir' to '$TargetDir'" ;
	if (!$DryRun) {
		Move-Item "$NewWebappDir" "$TargetDir" ; CheckLastExitStatus ;
		Push-RollbackCommand "Move-Item `"$TargetDir`" `"$NewWebappDir`" ; " ; 
	}

	# Log information
	$deploy_log_dir="$TargetDir/WEB-INF/data/logs" ;
	$deploy_log_file="$deploy_log_dir/deploy.log" ;
	if (!$DryRun) {
		$deploy_log_dir = New-Item -Type directory -Force $deploy_log_dir ; 
		if (Test-Path -PathType Leaf $deploy_log_file) {
			$deploy_log_file = Get-Item $deploy_log_file ; CheckLastExitStatus ;
		} else {		
			$deploy_log_file = New-Item -Type file -Force $deploy_log_file ; 
		}
		Write-Output $(Get-Date) " - Done deploying war '$WarPath'." | Out-File -Append "$deploy_log_file" ;
	}
	if ($LogMessage) {
		Write-Verbose "Append custom log message to '$deploy_log_file'" ;
		Write-Verbose "$LogMessage" ;
		if (!$DryRun) {
			Write-Output "$LogMessage" | Out-File -Append "$deploy_log_file" ;
		}
	}
	if ($log_diff) {
		Write-Verbose "Log diff between new and old webapp directories..."
		if (!$DryRun) {
			## TODO : diff -r --brief "$TargetDir" "$backup_webapp_path" | grep -i "docs/javadoc/"  | Out-File -Append "$deploy_log_file" ;
		}
	}
	if (!$DryRun) {
		Write-Output "`n`n`n" | Out-File -Append "$deploy_log_file" ;
	}

	# Cleanup backup and save all information in one place : the backuped webapp directory
	Write-Verbose "[webapp] Remove symlinks from backuped webapp" ;
	if (!$DryRun) {
		# cannot use remove-item which would follow symlink and remove diretory content...
		cmd /c "rmdir `"$backup_webapp_path/archives`"" ;
		cmd /c "rmdir `"$backup_webapp_path/upload`"" ;
		cmd /c "rmdir `"$backup_webapp_path/WEB-INF/data`"" ;
	}
	
	Write-Verbose "[webapp] Move backuped store, types and workflow in backuped webapp" ;
	if (!$DryRun) {
		$script:DeployBackupDir = "$backup_webapp_path/deploy-backup" ;
		if (Test-Path $DeployBackupDir) {
			$script:DeployBackupDir = Get-Item $DeployBackupDir ;
		} else {
			$script:DeployBackupDir = New-Item -Type directory -Force $DeployBackupDir ;
		}
		Write-Debug(" DeployBackupDir    : '$script:DeployBackupDir' - [$(Get-Type-FullName($script:DeployBackupDir))]") ;
	
		if ($backup_store_path -and (Test-Path -PathType Leaf $backup_store_path)) {
			Move-Item "$backup_store_path" "$DeployBackupDir/store.xml" ;
		}
		if ($backup_types_path -and (Test-Path -PathType Container $backup_types_path)) {
			Move-Item "$backup_types_path" "$DeployBackupDir/types" ;
		}
		if ($backup_workflows_path -and (Test-Path -PathType Container $backup_workflows_path)) {
			Move-Item "$backup_workflows_path" "$DeployBackupDir/workflows" ;
		}
	}

	Write-Output "DONE" ;
}

# cleanup temp directory
function Process-Cleanup() {
  Write-Verbose "Perform cleanup" ;
  if (!$DryRun -and (Test-Path "$TmpDir")) {
    Remove-Item  -Recurse -Force "$TmpDir" ;
  }
}

##############################################################
# Utility methods
##############################################################

function Get-Type-FullName($object) {
	if ($object) {
		return $object.GetType().FullName
	} else {
		return "null";
	}
}

##############################################################
# Main
##############################################################


function main() {
	if ($Verbose) {$VerbosePreference = "Continue"}
	if ($Debug) {$DebugPreference = "Continue"}
	
	if ($Help) {
		Write-Output "deploy.ps1 version $deployps1_revision $deployps1_date" ;
		Write-Usage ;
	}
	
	# Display TODO to developpers
	if ($Debug) {
		Select-String -pattern "^\s*## TODO" -path $MyInvocation.ScriptName | % {$_.Line};
	}
	
	# Remove all previous error to properly check them in CheckLastExitStatus
	$error.Clear();
		
	# Check arguments as well as environements 
	Write-Verbose("Checking requirements...");
	Check-Requirements;
	
	# Initialized required variables 
	Write-Verbose("Setup environement...");
	Setup-Environment;

	# Display Debugging information
	Write-Debug("-- Script parameter --") ;
	Write-Debug(" DryRun          : $script:DryRun");
	Write-Debug(" WarPath         : '$script:WarPath' - [$(Get-Type-FullName($script:WarPath))]");
	Write-Debug(" TargetDir       : '$script:TargetDir' - [$(Get-Type-FullName($script:TargetDir))]");
	Write-Debug(" StoreAction     : $script:StoreAction");
	Write-Debug(" BackupDirectory : '$script:BackupDirectory' - [$(Get-Type-FullName($script:BackupDirectory))]");
	Write-Debug("-- Working variables --") ;
	Write-Debug(" BackupSuffix    : $script:BackupSuffix") ;
	Write-Debug(" TmpDir          : '$script:TmpDir' - [$(Get-Type-FullName($script:TmpDir))]") ;
	Write-Debug(" JAVA_CLASSPATH  : $script:JAVA_CLASSPATH") ;
	Write-Debug(" RollbackScript    : '$script:RollbackScript' - [$(Get-Type-FullName($script:RollbackScript))]") ;
	Write-Debug(" RollbackScriptTmp : '$script:RollbackScriptTmp' - [$(Get-Type-FullName($script:RollbackScriptTmp))]") ;
	Write-Debug("-----------------------") ;
	
	# Start unzipping war and setting up symlink
	Unzip-War ;
	Copy-SymLinks ;
  
	# Update the content of the symlinked data according to options
	Process-Store ;
	Process-Types ;
	Process-Workflows ;
	
	# Replace the old webapp with the new updated one
	Process-Deploy

	# clean
	Process-Cleanup ; 
}		

main;
	